@@ -638,7 +638,7 @@ setMethod("geneSetSizes", signature("GsvaExprData"),
 #' [GMT file format specification](https://software.broadinstitute.org/cancer/software/gsea/wiki/index.php/Data_formats)
 #' but can nevertheless be encountered in the wild.
 #' The available choices are:
-#' * `first` (the default): drops all gene sets whose names are [`duplicated`]
+#' * `first` (the default): drops all gene sets whose names are duplicated
 #' according to the base R function and retains only the first occurence of a
 #' gene set name.
 #' * `drop`:  removes *all* gene sets that have a duplicated name, including its
@@ -892,7 +892,7 @@ geneIdsToGeneSetCollection <- function(geneIdsList,
 #' [GMT file format specification](https://software.broadinstitute.org/cancer/software/gsea/wiki/index.php/Data_formats)
 #' but can nevertheless be encountered in the wild.
 #' The available choices are:
-#' * `first` (the default): drops all gene sets whose names are [`duplicated`]
+#' * `first` (the default): drops all gene sets whose names are duplicated
 #' according to the base R function and retains only the first occurence of a
 #' gene set name.
 #' * `drop`:  removes *all* gene sets that have a duplicated name, including its
@@ -1035,6 +1035,16 @@ setMethod("unwrapData", signature("dgCMatrix"),
               return(container)
           })
 
+setMethod("unwrapData", signature("SVT_SparseArray"),
+          function(container, assay) {
+              return(container)
+          })
+
+setMethod("unwrapData", signature("DelayedMatrix"),
+          function(container, assay) {
+              return(container)
+          })
+
 setMethod("unwrapData", signature("ExpressionSet"),
           function(container, assay) {
               return(exprs(container))
@@ -1116,6 +1126,20 @@ setMethod("wrapData", signature(container="dgCMatrix"),
               return(dataMatrix)
           })
 
+setMethod("wrapData", signature(container="SVT_SparseArray"),
+          function(container, dataMatrix, geneSets) {
+              if (!missing(geneSets))
+                  attr(dataMatrix, "geneSets") <- geneSets
+              return(dataMatrix)
+          })
+
+setMethod("wrapData", signature(container="DelayedMatrix"),
+          function(container, dataMatrix, geneSets) {
+              if (!missing(geneSets))
+                  attr(dataMatrix, "geneSets") <- geneSets
+              return(dataMatrix)
+          })
+
 setMethod("wrapData", signature(container="ExpressionSet"),
           function(container, dataMatrix, geneSets) {
               rval <- new("ExpressionSet", exprs=dataMatrix,
